import collections


class _Path:
    def __init__(self, li, *, absolute):
        self.li = li
        self.absolute = absolute

        if self.absolute:
            assert len(self.li) > 0 and self.li[0].isRoot()

    def __str__(self):
        if len(self.li) < 2:
            res = ''
        else:
            res = []
            for _1, _2 in zip(self.li[:-1], self.li[1:]):
                assert _1.parent is _2 or _2.parent is _1
                res.append('..' if _1.parent is _2 else _2.name)
            res = '/'.join(res)

        if self.absolute:
            res = '/' + res
        if res == '':
            res = '.'
        return res


class _Node:
    def __init__(self, name):
        self.name = name
        self.parent = None
        self.children = []

    def isRoot(self):
        return self.parent is None

    def _abspath(self):
        li = []
        node = self
        while node is not None:
            li.insert(0, node)
            node = node.parent
        return _Path(li, absolute=True)

    def _relpathTo(self, dst):
        li_self = self._abspath().li
        li_dst = dst._abspath().li

        i = -1
        for _1, _2 in zip(li_self, li_dst):
            if _1 is not _2:
                break
            i += 1
        assert i >= 0  # because every abspath starts with the root node

        li = li_self[-1:i:-1] + li_dst[i:]
        return _Path(li, absolute=False)

    def _relpathFrom(self, src):
        return src._relpathTo(self)

    def abspath(self):
        return str(self._abspath())

    def relpathTo(self, dst):
        return str(self._relpathTo(dst))

    def relpathFrom(self, src):
        return str(self._relpathFrom(src))

    def _append(self, node):
        node.parent = self
        self.children.append(node)

    def __str__(self):
        return self.abspath()


class ExploringTree:
    def __init__(self):
        self._root = _Node('')
        self._whereami = self._root
        self._old_whereami_stack = []

    def whereami(self):
        return self._whereami

    def up(self):
        assert self._whereami is not self._root

        self._whereami = self._whereami.parent

    def down(self, name):
        assert name != ''

        d = {_.name: _ for _ in self._whereami.children}
        if name in d:
            self._whereami = d[name]
        else:
            node = _Node(name)
            self._whereami._append(node)
            self._whereami = node

    def jump(self, dst):
        if isinstance(dst, _Node):
            self._whereami = dst
            return dst
        assert isinstance(dst, str)

        if len(dst) == 0:
            return self._whereami

        li = dst.split('/')
        if li[0] == '':
            self._whereami = self._root
            li = li[1:]

        for name in li:
            if name == '.':
                continue
            elif name == '..':
                self.up()
            elif name == '':
                continue
            else:
                self.down(name)

        return self._whereami

    def see(self, dst):
        with self:
            return self.jump(dst)

    def abspath(self):
        return self._whereami.abspath()

    def relpathTo(self, dst):
        return self._whereami.relpathTo(dst)

    def relpathFrom(self, src):
        return self._whereami.relpathFrom(src)

    def __enter__(self):
        self._old_whereami_stack.append(self._whereami)

    def __exit__(self, exc_type, exc_value, traceback):
        self._whereami = self._old_whereami_stack.pop()



import json
import collections


class _MyDictJSONEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, MyDict):
            return o.data
        return super().default(o)


class MyDict(collections.UserDict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def to_json(self):
        return _MyDictJSONEncoder(indent=4).encode(self)



import subprocess
import os
import sys
import logging
import shutil
import re
import xmlrpc.client

import requests
import jinja2

from .resource import load_resource


def _esc(s):
    return s.replace('\\', '\\\\').replace('"', '\\"')


class DownloaderCanNotWork(Exception):
    pass


_url_pattern_github = re.compile(r'https://github.com/([^/]+)/([^/]+)/blob/(.+)')


def _fix_url(url):
    m = _url_pattern_github.fullmatch(url)
    if m:
        return f'https://raw.githubusercontent.com/{m.group(1)}/{m.group(2)}/refs/heads/{m.group(3)}'
    return url


class Downloader:
    def __init__(self, *, dl_tasks):
        self._dl_tasks = dl_tasks

        for _ in self._dl_tasks:
            _['url'] = _fix_url(_['url'])

    def download(self):
        pass

    def check(self):
        pass


class DownloaderTS(Downloader):
    def __init__(self, *args, ts, **kwargs):
        super().__init__(*args, **kwargs)

        self._ts = ts
        self._dl = ts.register_task(self._dl, ttl=3, desc='download')

    def _dl(self, *, url, filename):
        pass

    def download(self):
        self.check()

        for _ in self._dl_tasks:
            self._dl(url=_['url'], filename=_['filename'])

        failures = self._ts.wait()
        return [_1.kwargs for _1, _2 in failures]


class DownloaderBuiltin(DownloaderTS):
    def _dl(self, *, url, filename):
        os.makedirs(os.path.dirname(filename), exist_ok=True)

        with requests.get(url, stream=True) as r:
            r.raise_for_status()
            r.raw.decode_content = True
            with open(filename, 'wb') as ofs:
                shutil.copyfileobj(r.raw, ofs)


class CheckExeMixIn:
    def _exe(self):
        pass

    def check(self):
        exe = self._exe()
        try:
            subprocess.run(
                [exe, '--version'],
                stdout=subprocess.DEVNULL,
                stderr=subprocess.STDOUT,
                check=True,
            )
        except OSError:
            raise DownloaderCanNotWork(
                'The executable %s is not installed or not in PATH' % exe
            )


class DownloaderSubprocess(CheckExeMixIn, DownloaderTS):
    def _dl(self, *, url, filename):
        subprocess.run(
            self._cmdline(url, filename),
            stdout=subprocess.DEVNULL,
            stderr=subprocess.STDOUT,
            check=True,
        )

    def _cmdline(self, url, filename):
        pass


class DownloaderCurl(DownloaderSubprocess):
    def _cmdline(self, url, filename):
        return [
            'curl',
            '--create-dirs',
            '--globoff',
            '--url',
            url,
            '--output',
            filename,
        ]

    def _exe(self):
        return 'curl'


class DownloaderAria2(DownloaderSubprocess):
    def _cmdline(self, url, filename):
        return [
            'aria2c',
            '--allow-overwrite=true',
            '--always-resume=true',
            '--auto-file-renaming=false',
            '--continue=true',
            '-d',
            os.path.dirname(filename),
            '-o',
            os.path.basename(filename),
            url,
        ]

    def _exe(self):
        return 'aria2c'


class DownloaderInput_file(Downloader):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._env = jinja2.Environment()
        self._env.filters['_dirname'] = lambda _: os.path.dirname(_)
        self._env.filters['_basename'] = lambda _: os.path.basename(_)
        self._env.filters['_esc'] = _esc

    def download(self):
        template = self._env.from_string(
            load_resource(self._template_path).decode('UTF-8')
        )
        return template.render(dl_tasks=self._dl_tasks)

    @property
    def _template_path(self):
        pass


class DownloaderCurl_input_file(DownloaderInput_file):
    @property
    def _template_path(self):
        return 'template/curl_input_file'


class DownloaderAria2_input_file(DownloaderInput_file):
    @property
    def _template_path(self):
        return 'template/aria2_input_file'


class DownloaderUget(CheckExeMixIn, Downloader):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def download(self):
        self.check()

        for _ in self._dl_tasks:
            _url = _['url']
            _filename = _['filename']
            cmdline = [
                self._exe(),
                '--quiet',
                '--folder=%s' % os.path.dirname(_filename),
                '--filename=%s' % os.path.basename(_filename),
                _url,
            ]
            subprocess.run(
                cmdline, stdout=subprocess.DEVNULL, stderr=subprocess.STDOUT, check=True
            )

            logging.info(
                'add downloading task: url=%s, filename=%s' % (_url, _filename)
            )

    def _exe(self):
        return 'uget' if sys.platform == 'win32' else 'uget-gtk'


class DownloaderAria2_rpc(Downloader):
    def __init__(self, *args, url=None, secret=None, **kwargs):
        super().__init__(*args, **kwargs)

        self._client = xmlrpc.client.ServerProxy(url)
        self._secret = secret

    def download(self):
        self.check()

        for _ in self._dl_tasks:
            _url = _['url']
            _filename = _['filename']
            _args = [
                [_url],
                {
                    'dir': os.path.dirname(_filename),
                    'out': os.path.basename(_filename),
                    'allow-overwrite': 'true',
                    'always-resume': 'true',
                    'auto-file-renaming': 'false',
                    'continue': 'true',
                },
            ]
            if self._secret is not None:
                _args.insert(0, 'token:%s' % self._secret)

            self._client.aria2.addUri(*_args)
            logging.info(
                'add downloading task: url=%s, filename=%s' % (_url, _filename)
            )

    def check(self):
        try:
            self._client.system.listMethods()
        except ConnectionError:
            raise DownloaderCanNotWork('fail to connnect to the XML-RPC server')



import logging
import argparse
import os
import pickle
import json
import sys
import textwrap

import requests

from tqdm import tqdm

import dl_coursera

from dl_coursera.lib.misc import change_ext, get_latest_app_version
from dl_coursera.lib.TaskScheduler import TaskScheduler
from dl_coursera.Crawler import Crawler
from dl_coursera.DLTaskGatherer import DLTaskGatherer
from dl_coursera.Downloader import DownloaderBuiltin
from dl_coursera.define import *


def _dir_cache(outdir, slug):
    return os.path.join(outdir, slug, '.cache')


def _file_log(outdir, slug):
    return os.path.join(_dir_cache(outdir, slug), 'main.log')


def _file_pkl_crawl(outdir, slug):
    return os.path.join(_dir_cache(outdir, slug), 'crawl.pkl')


def _file_json_gather(outdir, slug):
    return os.path.join(_dir_cache(outdir, slug), 'gather.json')


def _file_json_download_dl_tasks_failed(outdir, slug):
    return os.path.join(_dir_cache(outdir, slug), 'download.dl_tasks_failed.json')


def crawl(cookies_file, slug, outdir):
    file_pkl = _file_pkl_crawl(outdir, slug)
    if os.path.exists(file_pkl):
        with open(file_pkl, 'rb') as ifs:
            return pickle.load(ifs)

    with requests.Session() as sess:
        Crawler._login(sess, cookies_file=cookies_file)

        # Check whether the specialization/course exists

        if 'elements' in sess.get(URL_SPEC(slug)).json():
            isSpec = True
        elif 'elements' in sess.get(URL_COURSE_1(slug)).json():
            isSpec = False
        else:
            raise NotFoundExcepton(slug)

        # Check whether the cookies_file expires

        course = Course(slug=COURSE_0)
        d = sess.get(URL_COURSE_1(course['slug'])).json()
        course['id'] = d['elements'][0]['id']

        d = sess.get(URL_COURSE_REFERENCES(course['id'])).json()
        if d.get('errorCode') == 'Not Authorized':
            raise CookiesExpiredException()
        assert 'errorCode' not in d

    with TaskScheduler() as ts, requests.Session() as sess:
        with tqdm(
            desc='Crawling...',
            bar_format='{bar:31} [{percentage:3.0f}%] {n_fmt}/{total_fmt} {desc}',
        ) as bar:
            total = 0
            done = 0

            def _hook_add():
                nonlocal total
                total += 1
                bar.reset(total)
                bar.update(done)
                bar.refresh()

            def _hook_done():
                nonlocal done
                done += 1
                bar.update()
                bar.refresh()

            def _hook_retry():
                bar.refresh()

            ts.start(
                n_worker=1,
                hook_add=_hook_add,
                hook_done=_hook_done,
                hook_retry=_hook_retry,
            )
            crawler = Crawler(ts=ts, sess=sess, cookies_file=cookies_file)
            soc = crawler.crawl(slug=slug, isSpec=isSpec)

    with open(file_pkl, 'wb') as ofs:
        pickle.dump(soc, ofs)

    file_json = change_ext(file_pkl, 'json')
    with open(file_json, 'w', encoding='UTF-8') as ofs:
        ofs.write(soc.to_json())

    return soc


def gather_dl_tasks(outdir, soc):
    file_json = _file_json_gather(outdir, soc['slug'])
    if os.path.exists(file_json):
        with open(file_json, encoding='UTF-8') as ifs:
            return json.load(ifs)

    dl_tasks = DLTaskGatherer(soc=soc, outdir=outdir).gather()
    with open(file_json, 'w', encoding='UTF-8') as ofs:
        json.dump(dl_tasks, ofs, indent=4)

    return dl_tasks


def download(dl_tasks, slug, outdir):
    file_json = _file_json_download_dl_tasks_failed(outdir, slug)
    if os.path.exists(file_json):
        with open(file_json, encoding='UTF-8') as ifs:
            dl_tasks = json.load(ifs)

    if len(dl_tasks) == 0:
        return

    with TaskScheduler() as ts:
        with tqdm(
            desc='Downloading...',
            bar_format='{bar:31} [{percentage:3.0f}%] {n_fmt}/{total_fmt} {desc}',
            total=len(dl_tasks),
        ) as bar:

            def _hook_done():
                bar.update(1)
                bar.refresh()

            def _hook_retry():
                bar.refresh()

            ts.start(n_worker=1, hook_done=_hook_done, hook_retry=_hook_retry)
            _cls_downloader = DownloaderBuiltin
            dl_tasks_failed = _cls_downloader(dl_tasks=dl_tasks, ts=ts).download()

    with open(file_json, 'w', encoding='UTF-8') as ofs:
        json.dump(dl_tasks_failed, ofs, indent=4)


def config_logger(logfile: str):
    logger = logging.getLogger()
    for _ in list(logger.handlers):
        logger.removeHandler(_)
    logger.setLevel(logging.NOTSET)

    stderrHandler = logging.StreamHandler()
    fileHandler = logging.FileHandler(filename=logfile, encoding='UTF-8', mode='w')

    logger.addHandler(stderrHandler)
    logger.addHandler(fileHandler)

    formatter = logging.Formatter(
        fmt='%(asctime)s - %(levelname)s - %(message)s',
        style='%',
        datefmt='%Y-%m-%d %H:%M:%S',
    )

    def _config_handler(h: logging.Handler, name: str, level: int):
        h.set_name(name)
        h.setFormatter(formatter)
        h.setLevel(level)

    _config_handler(stderrHandler, 'stderr', logging.ERROR)
    _config_handler(fileHandler, 'file', logging.INFO)


def main():
    parser = argparse.ArgumentParser(
        allow_abbrev=False,
        add_help=True,
        description='A simple, fast, and reliable Coursera crawling & downloading tool',
        epilog=textwrap.dedent(
            """
            If the command succeeds, you shall see `Done :-)`.
            If errors occur, visit `https://github.com/FLZ101/dl_coursera`
            for the troubleshooting guide.
            """
        ),
    )
    parser.add_argument('--cookies', required=True, help='path of the cookies file')
    parser.add_argument(
        '--outdir', default='.', help="the output directory. Default: `.'"
    )
    parser.add_argument(
        '--version', action='version', version='%%(prog)s %s' % dl_coursera.app_version
    )
    parser.add_argument('slug', help='slug of the specialization/course')

    args = vars(parser.parse_args())

    latest_version = get_latest_app_version()
    if latest_version > dl_coursera.app_version:
        msg = textwrap.dedent(
            f"A newer version {latest_version} is available.",
            file=sys.stderr,
        )
        print(msg, file=sys.stderr, flush=True)

    outdir = args['outdir']
    slug = args['slug']
    os.makedirs(_dir_cache(outdir, slug), exist_ok=True)

    config_logger(_file_log(outdir, slug))

    soc = crawl(args['cookies'], slug, outdir)

    dl_tasks = gather_dl_tasks(outdir, soc)

    download(dl_tasks, slug, outdir)

    sys.stderr.flush()
    print('Done :-)')


if __name__ == '__main__':
    main()



import os
import logging
import base64

from http.cookiejar import MozillaCookieJar

from .define import *

from .lib.misc import format_dict, TmpFile
from .markup import CML

PRIO_SPEC = 'A'
PRIO_COURSE = 'B'
PRIO_COURSE_MATERIAL = 'C'


class Crawler:
    @staticmethod
    def _login(sess, cookies_file=None, cookies_base64=None):
        if cookies_file is None:
            if cookies_base64 is None:
                # the env should contain $(base64 -w 0 cookies.txt)
                cookies_base64 = os.environ.get('DL_COURSERA_COOKIES_BASE64')
                assert cookies_base64

            cookies = base64.standard_b64decode(cookies_base64)

            with TmpFile() as tmpfile:
                with open(tmpfile, 'wb') as ofs:
                    ofs.write(cookies)

                cj = MozillaCookieJar()
                cj.load(tmpfile)
        else:
            cj = MozillaCookieJar()
            cj.load(cookies_file)

        sess.cookies.update(cj)

    @staticmethod
    def _get(sess, url):
        resp = sess.get(url)
        d = resp.json()
        if 'errorCode' in d:
            raise BadResponseException(d)
        return d

    def __init__(self, *, ts, sess, cookies_file=None):
        self._ts = ts
        self._sess = sess
        self._cookies_file = cookies_file

        self._loggedin = False

        def attach(func):
            setattr(self, func.__name__, func)
            return func

        @attach
        @ts.register_task(
            priority=PRIO_SPEC,
            ttl=3,
            format_kwargs=lambda _: format_dict({'spec': _['spec']['slug']}),
        )
        def crawl_spec(*, spec):
            d = Crawler._get(sess, URL_SPEC(spec['slug']))
            if 'elements' not in d:
                raise SpecNotExistExcepton(spec['slug'])

            _ = d['elements'][0]
            spec['id'] = _['id']
            spec['name'] = _['name']

            assert spec['slug'] == _['slug']

            for id_ in _['courseIds']:
                spec['courses'].append(Course(id_=id_))

            id2slug = {}
            for _ in d['linked']['courses.v1']:
                id2slug[_['id']] = _['slug']

            for _ in spec['courses']:
                _['slug'] = id2slug[_['id']]

            logging.info(
                'Courses of specialization %s: %s'
                % (spec['slug'], ', '.join([_['slug'] for _ in spec['courses']]))
            )

            for _ in spec['courses']:
                crawl_course(course=_)

        @attach
        @ts.register_task(
            priority=PRIO_COURSE,
            ttl=3,
            format_kwargs=lambda _: format_dict({'cource': _['course']['slug']}),
        )
        def crawl_course(*, course):
            d = Crawler._get(sess, URL_COURSE_1(course['slug']))
            if 'elements' not in d:
                raise CourseNotExistExcepton(course['slug'])

            _ = d['elements'][0]
            course['name'] = _['name']
            course['id'] = _['id']

            assert course['slug'] == _['slug']

            # ------

            d = Crawler._get(sess, URL_COURSE_2(course['slug']))['linked']

            id2item = {}
            for _ in d['onDemandCourseMaterialItems.v2']:
                typeName = _['contentSummary']['typeName']
                if typeName in [
                    'exam',
                    'quiz',
                    'phasedPeer',
                    'discussionPrompt',
                    'gradedProgramming',
                    'programming',
                ]:
                    continue

                if typeName not in ['lecture', 'notebook', 'supplement']:
                    logging.warning(
                        '[crawl_course] unknown typeName=%s\n%s' % (typeName, _)
                    )
                    continue

                if _.get('isLocked'):
                    logging.info('[crawl_course] locked item: %s' % _)
                    continue

                if typeName == 'lecture':
                    id2item[_['id']] = CourseMaterialLecture(
                        id_=_['id'], name=_['name'], slug=_['slug']
                    )
                elif typeName == 'notebook':
                    id2item[_['id']] = CourseMaterialNotebook(
                        id_=_['id'], name=_['name'], slug=_['slug']
                    )
                elif typeName == 'supplement':
                    id2item[_['id']] = CourseMaterialSupplement(
                        id_=_['id'], name=_['name'], slug=_['slug']
                    )

            id2lesson = {}
            for _ in d['onDemandCourseMaterialLessons.v1']:
                lesson = CourseMaterialLesson(
                    id_=_['id'], name=_['name'], slug=_['slug']
                )
                for id_item in _['itemIds']:
                    item = id2item.get(id_item)
                    if item is not None:
                        lesson['items'].append(item)

                if len(lesson['items']) > 0:
                    id2lesson[lesson['id']] = lesson

            for _ in d['onDemandCourseMaterialModules.v1']:
                module = CourseMaterialModule(
                    id_=_['id'], name=_['name'], slug=_['slug']
                )

                for id_ in _['lessonIds']:
                    lesson = id2lesson.get(id_)
                    if lesson is not None:
                        module['lessons'].append(lesson)

                if len(module['lessons']) > 0:
                    course['modules'].append(module)

            # ------

            crawl_course_references(course=course)

            for module in course['modules']:
                for lesson in module['lessons']:
                    for item in lesson['items']:
                        if item['type'] == 'Lecture':
                            crawl_lecture(course=course, lecture=item)
                        elif item['type'] == 'Supplement':
                            crawl_supplement(course=course, supplement=item)

        def _crawl_course_ref(course, id_ref=None):
            if not id_ref:
                d = Crawler._get(sess, URL_COURSE_REFERENCES(course['id']))
            else:
                d = Crawler._get(sess, URL_COURSE_REFERENCE(course['id'], id_ref))

            itemId2ref = {}
            for _ in d['elements']:
                ref = CourseReference(id_=_['shortId'], name=_['name'], slug=_['slug'])
                course['references'].append(ref)

                itemId = _['content'][
                    'org.coursera.ondemand.reference.AssetReferenceContent'
                ]['assetId']
                itemId2ref[itemId] = ref

            for _ in d['linked']['openCourseAssets.v1']:
                typeName = _['typeName']
                if typeName == 'cml':
                    cml = CML(_['definition']['value'])
                    assets, assetIDs, refids = cml.get_resources()
                    assets += crawl_assets(assetIDs)
                    html = cml.to_html(assets=assets)

                    itemId2ref[_['id']]['item'] = CourseMaterialSupplementItemCML(
                        html=html, assets=assets
                    )

                    for refid in refids:
                        crawl_course_reference(course=course, id_ref=refid)
                else:
                    logging.warning(
                        "[_crawl_course_ref] unknown typeName=%s\n%s" % (typeName, _)
                    )

        @ts.register_task(
            priority=PRIO_COURSE_MATERIAL,
            ttl=3,
            format_kwargs=lambda _: format_dict({'cource': _['course']['slug']}),
        )
        def crawl_course_references(*, course):
            _crawl_course_ref(course)

        @ts.register_task(
            priority=PRIO_COURSE_MATERIAL,
            ttl=3,
            format_kwargs=lambda _: format_dict(
                {'cource': _['course']['slug'], 'id_ref': _['id_ref']}
            ),
        )
        def crawl_course_reference(*, course, id_ref):
            for ref in course['references']:
                if id_ref == ref['id']:
                    return
            _crawl_course_ref(course, id_ref)

        @ts.register_task(
            priority=PRIO_COURSE_MATERIAL,
            ttl=3,
            format_kwargs=lambda _: format_dict(
                {'course': _['course']['slug'], 'lecture': _['lecture']['slug']}
            ),
        )
        def crawl_lecture(*, course, lecture):
            # lecture videos
            d = Crawler._get(sess, URL_LECTURE_1(course['id'], lecture['id']))

            for _ in d['linked']['onDemandVideos.v1']:
                url_subtitle = _['subtitles'].get('en')
                if url_subtitle is not None:
                    url_subtitle = URL_ROOT + url_subtitle

                _ = _['sources']['byResolution']
                # choose the video with highest resolution
                url_video = _[sorted(_.keys())[-1]]
                url_video = url_video['mp4VideoUrl']

                lecture['videos'].append(
                    Video(url_video=url_video, url_subtitle=url_subtitle)
                )

            # lecture assets
            d = Crawler._get(sess, URL_LECTURE_2(course['id'], lecture['id']))

            assets = []
            assetIDs = []
            for _ in d['linked']['openCourseAssets.v1']:
                typeName = _['typeName']
                if typeName == 'asset':
                    assetIDs.append(_['definition']['assetId'])
                elif typeName == 'url':
                    assets.append(
                        Asset(
                            id_=_['id'],
                            url=_['definition']['url'],
                            name=_['definition']['name'],
                        )
                    )
                else:
                    logging.warning(
                        "[crawl_lecture] unknown typeName=%s\n%s" % (typeName, _)
                    )

            assets += crawl_assets(assetIDs)
            lecture['assets'] = assets

        @ts.register_task(
            priority=PRIO_COURSE_MATERIAL,
            ttl=3,
            format_kwargs=lambda _: format_dict(
                {'course': _['course']['slug'], 'supplement': _['supplement']['slug']}
            ),
        )
        def crawl_supplement(course, supplement):
            d = Crawler._get(sess, URL_SUPPLEMENT(course['id'], supplement['id']))
            for _ in d['linked']['openCourseAssets.v1']:
                typeName = _['typeName']
                if typeName == 'cml':
                    cml = CML(_['definition']['value'])
                    assets, assetIDs, refids = cml.get_resources()
                    assets += crawl_assets(assetIDs)
                    html = cml.to_html(assets=assets)

                    supplement['items'].append(
                        CourseMaterialSupplementItemCML(html=html, assets=assets)
                    )

                    for refid in refids:
                        crawl_course_reference(course=course, id_ref=refid)
                else:
                    logging.warning(
                        "[crawl_supplement] unknown typeName=%s\n%s" % (typeName, _)
                    )

        def crawl_assets(ids):
            if len(ids) == 0:
                return []

            d = Crawler._get(sess, URL_ASSET(ids))

            assets = []
            for _ in d['elements']:
                id_ = _['id']
                url = _['url']['url']
                name = _asset_name(_['name'], _['fileExtension'])
                assets.append(Asset(id_=id_, url=url, name=name))

            if len(assets) != len(ids):
                logging.warning(
                    '[crawl_assets] unexpected number of assets\n%s\n%s' % (ids, assets)
                )
            return assets

        def _asset_name(name, fileExtension):
            fileExtension = '.' + fileExtension
            if not name.endswith(fileExtension):
                name += fileExtension
            return name

    def login(self):
        if not self._loggedin:
            Crawler._login(self._sess, self._cookies_file)
            self._loggedin = True

    def crawl(self, *, slug, isSpec):
        if not self._loggedin:
            self.login()

        if isSpec:
            res = Spec(slug=slug)
            self.crawl_spec(spec=res)
        else:
            res = Course(slug=slug)
            self.crawl_course(course=res)

        self._ts.wait()
        return res



import os
import tempfile


def format_dict(d):
    return ', '.join(['%s=%s' % (k, v) for k, v in d.items()])


def change_ext(filename, ext):
    return os.path.splitext(filename)[0] + ('' if ext == '' else '.' + ext)


def url_basename(url, *, _d={}):
    import re
    from urllib.parse import urlparse

    if _d.get('pat') is None:
        _d['pat'] = re.compile(r'([^/]*)$')

    match = _d['pat'].search(urlparse(url).path)
    return match.group(1)


class TmpFile:
    def __init__(self, suffix=None, prefix=None, dir=None):
        self._suffix = suffix
        self._prefix = prefix
        self._dir = dir

    def __enter__(self):
        fd, filename = tempfile.mkstemp(self._suffix, self._prefix, self._dir)
        os.close(fd)

        self._filename = filename
        return filename

    def __exit__(self, exc_type, exc_value, traceback):
        # os.remove(self._filename)
        pass


def get_latest_app_version():
    import requests

    resp = requests.get('https://pypi.org/pypi/dl-coursera/json')
    d = resp.json()

    return sorted(d['releases'].keys())[-1]



import logging
import copy
import re

from urllib.parse import quote

import bs4

import jinja2

from .resource import load_resource
from .define import URL_ROOT


def _is_root(e):
    return e.parent is None


def _is_tag(e):
    return isinstance(e, bs4.Tag)


def _has_no_child(tag):
    return len(tag.contents) == 0


class Traversal:
    '''depth-first traversal'''

    def __init__(self, root, *, tagOnly=False):
        assert _is_root(root)

        self._e = root
        self._skip_children = False
        self._tagOnly = tagOnly

    def __iter__(self):
        return self

    def __next__(self):
        e = self._next()
        if e is None:
            raise StopIteration('no more element')

        self._e = e
        self._skip_children = False
        return e

    def _next(self):
        if (not self._skip_children) and _is_tag(self._e):
            for e in self._e.children:
                if (not self._tagOnly) or _is_tag(e):
                    return e

        e = self._e
        while True:
            if _is_root(e):
                return None

            for _e in e.next_siblings:
                if (not self._tagOnly) or _is_tag(_e):
                    return _e

            e = e.parent

    def skip_children(self):
        self._skip_children = True


class CML:
    def __init__(self, doc):
        doc = doc.translate(doc.maketrans('\u21b5', ' '))
        self._root = bs4.BeautifulSoup(doc, 'lxml-xml')

        self._assets = None
        self._assetIDs = None
        self._refids = None

        self._html = None

    def get_resources(
        self,
        *,
        _pat_ref=re.compile(r'%s/learn/[^/]+/resources/([0-9a-zA-Z-]+)' % URL_ROOT)
    ):
        if self._assets is None:
            self._assets = []
            self._assetIDs = []
            self._refids = []

            for e in Traversal(self._root, tagOnly=True):
                if e.name == 'asset':
                    self._assetIDs.append(e['id'])
                elif e.name == 'img':
                    if e.get('src'):
                        import uuid

                        from .lib.misc import url_basename
                        from .define import Asset

                        id_ = str(uuid.uuid4())
                        e['assetId'] = id_
                        url = e['src']
                        name = url_basename(url)
                        self._assets.append(Asset(id_=id_, url=url, name=name))
                    else:
                        self._assetIDs.append(e['assetId'])
                elif e.name == 'a':
                    match = _pat_ref.match(e['href'])
                    if match:
                        _ = match.group(1)
                        self._refids.append(_)
                        e['refid'] = _

        return self._assets, self._assetIDs, self._refids

    def to_html(self, *, assets):
        if self._html is not None:
            return self._html

        asset_by_id = {_['id']: _ for _ in assets}

        def _assetName(id_):
            return asset_by_id[id_]['name']

        html = bs4.BeautifulSoup('', 'lxml')
        d = {}

        def _add(e0, e1):
            parent1 = html
            _e = e0
            while _e is not None:
                if id(_e) in d:
                    parent1 = d[id(_e)]
                    break
                _e = _e.parent

            if (parent1 is html) and (not _is_tag(e0)):
                return

            if _is_tag(e0):
                d[id(e0)] = e1
            parent1.append(e1)

        tr = Traversal(self._root)
        for e0 in tr:
            if isinstance(e0, bs4.NavigableString):
                _li = str(e0).split('$$')
                hasMath = False
                for _ in _li:
                    if not hasMath:
                        _add(e0, _)
                    else:
                        _span = bs4.Tag(name='span')
                        _span['hasMath'] = 'true'
                        _span.append(_)
                        _add(e0, _span)

                    hasMath = not hasMath

                continue

            if not _is_tag(e0):
                continue

            if e0.name == 'asset':
                assert _has_no_child(e0)

                e1 = bs4.Tag(name='p')
                e1['class'] = 'asset'
                e1.append(_assetName(e0['id']))

            elif e0.name == 'img':
                assert _has_no_child(e0)

                e1 = bs4.Tag(name='img')
                e1['src'] = e1['alt'] = _assetName(e0['assetId'])
                e1['src'] = quote(e1['src'])

            elif e0.name == 'heading':
                e1 = bs4.Tag(name='h%d' % int(e0['level']))

            elif e0.name == 'text':
                e1 = bs4.Tag(name='p')

            elif e0.name == 'list':
                bulletType = e0['bulletType']
                if bulletType == 'numbers':
                    e1 = bs4.Tag(name='ol')
                    e1['type'] = '1'
                elif bulletType == 'bullets':
                    e1 = bs4.Tag(name='ul')
                else:
                    e1 = bs4.Tag(name='ul')
                    logging.warning('[CML] unknown bulletType=%s' % bulletType)

            elif e0.name == 'a':
                e1 = bs4.Tag(name='a')
                e1['href'] = e0['href']
                if e0.get('refid'):
                    e1['refid'] = e0['refid']

            elif e0.name == 'code':
                e1 = bs4.Tag(name='pre')
                e1.append(copy.copy(e0))

                tr.skip_children()

            elif e0.name in [
                'li',
                'strong',
                'em',
                'u',
                'table',
                'tr',
                'td',
                'th',
                'sup',
                'sub',
            ]:
                e1 = bs4.Tag(name=e0.name)

            elif e0.name in ['co-content']:
                continue

            else:
                logging.warning('[CML] unknown e0.name=%s\n%s' % (e0.name, e0))
                continue

            _add(e0, e1)

        self._html = str(html)
        return self._html


def render_supplement(*, content, resource_path, title='', __={}):
    if __.get('template') is None:
        __['template'] = jinja2.Template(
            load_resource('template/supplement.html').decode('UTF-8')
        )
    return __['template'].render(
        content=content, resource_path=resource_path, title=title
    )



import unittest


class TestMisc(unittest.TestCase):
    def test_pypi_api(self):
        from dl_coursera.lib.misc import get_latest_app_version

        ver = get_latest_app_version()
        self.assertRegex(ver, r'\d+\.\d+\.\d+')



import setuptools

import dl_coursera


def _read(filename):
    with open(filename, encoding='UTF-8') as ifs:
        return ifs.read()


def _readlines(filename):
    return [_ for _ in _read(filename).split('\n') if len(_) > 0]


setuptools.setup(
    name=dl_coursera.app_name,
    version=dl_coursera.app_version,
    author='fengleizZZ',
    author_email='fenglei4518@hotmail.com',
    description='A simple, fast, and reliable Coursera crawling & downloading tool',
    long_description=_read('README.md'),
    long_description_content_type='text/markdown',
    url='https://github.com/FLZ101/dl_coursera',
    keywords=[
        'dl_coursera',
        'coursera',
        'education',
        'MOOC',
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: End Users/Desktop',
        'Programming Language :: Python :: 3 :: Only',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Topic :: Education',
    ],
    python_requires='>=3.8',
    install_requires=_readlines('requirements.txt'),
    extras_require={'dev': _readlines('requirements-dev.txt')},
    py_modules=['dl_coursera_run'],
    packages=setuptools.find_packages(),
    entry_points={
        'console_scripts': ['%s=dl_coursera_run:main' % dl_coursera.app_name]
    },
    include_package_data=True,
    package_data={'dl_coursera.resource': ['*.zip', 'template/*']},
    platforms=['any'],
)



import json

from .lib.MyDict import MyDict

URL_ROOT = 'https://www.coursera.org'
COURSE_0 = 'learning-how-to-learn'


def URL_SPEC(slug):
    return (
        URL_ROOT
        + '/api/onDemandSpecializations.v1?q=slug&slug={}&fields=courseIds,interchangeableCourseIds,launchedAt,logo,memberships,metadata,partnerIds,premiumExperienceVariant,onDemandSpecializationMemberships.v1(suggestedSessionSchedule),onDemandSpecializationSuggestedSchedule.v1(suggestedSessions),partners.v1(homeLink,name),courses.v1(courseProgress,description,membershipIds,startDate,v2Details,vcMembershipIds),v2Details.v1(onDemandSessions,plannedLaunchDate),memberships.v1(grade,vcMembershipId),vcMemberships.v1(certificateCodeWithGrade)&includes=courseIds,memberships,partnerIds,onDemandSpecializationMemberships.v1(suggestedSessionSchedule),courses.v1(courseProgress,membershipIds,v2Details,vcMembershipIds),v2Details.v1(onDemandSessions)'.format(
            slug
        )
    )


def URL_COURSE_1(slug):
    return (
        URL_ROOT
        + '/api/onDemandCourses.v1?q=slug&slug={}&includes=instructorIds%2CpartnerIds%2C_links&fields=brandingImage%2CcertificatePurchaseEnabledAt%2Cpartners.v1(squareLogo%2CrectangularLogo)%2Cinstructors.v1(fullName)%2CoverridePartnerLogos%2CsessionsEnabledAt%2CdomainTypes%2CpremiumExperienceVariant%2CisRestrictedMembership'.format(
            slug
        )
    )


def URL_COURSE_2(slug):
    return (
        URL_ROOT
        + '/api/onDemandCourseMaterials.v2/?q=slug&slug={}&includes=modules%2Clessons%2CpassableItemGroups%2CpassableItemGroupChoices%2CpassableLessonElements%2Citems%2Ctracks%2CgradePolicy&fields=moduleIds%2ConDemandCourseMaterialModules.v1(name%2Cslug%2Cdescription%2CtimeCommitment%2ClessonIds%2Coptional%2ClearningObjectives)%2ConDemandCourseMaterialLessons.v1(name%2Cslug%2CtimeCommitment%2CelementIds%2Coptional%2CtrackId)%2ConDemandCourseMaterialPassableItemGroups.v1(requiredPassedCount%2CpassableItemGroupChoiceIds%2CtrackId)%2ConDemandCourseMaterialPassableItemGroupChoices.v1(name%2Cdescription%2CitemIds)%2ConDemandCourseMaterialPassableLessonElements.v1(gradingWeight%2CisRequiredForPassing)%2ConDemandCourseMaterialItems.v2(name%2Cslug%2CtimeCommitment%2CcontentSummary%2CisLocked%2ClockableByItem%2CitemLockedReasonCode%2CtrackId%2ClockedStatus%2CitemLockSummary)%2ConDemandCourseMaterialTracks.v1(passablesCount)&showLockedItems=true'.format(
            slug
        )
    )


def URL_COURSE_REFERENCES(id_course):
    return (
        URL_ROOT
        + '/api/onDemandReferences.v1/?courseId={}&q=courseListed&fields=name%2CshortId%2Cslug%2Ccontent&includes=assets'.format(
            id_course
        )
    )


def URL_COURSE_REFERENCE(id_course, id_ref):
    return (
        URL_ROOT
        + '/api/onDemandReferences.v1/?courseId={}&q=shortId&shortId={}&fields=name%2CshortId%2Cslug%2Ccontent&includes=assets'.format(
            id_course, id_ref
        )
    )


def URL_LECTURE_1(id_course, id_lecture):
    return (
        URL_ROOT
        + '/api/onDemandLectureVideos.v1/{}~{}?includes=video&fields=onDemandVideos.v1(sources%2Csubtitles%2CsubtitlesVtt%2CsubtitlesTxt)'.format(
            id_course, id_lecture
        )
    )


def URL_LECTURE_2(id_course, id_lecture):
    return (
        URL_ROOT
        + '/api/onDemandLectureAssets.v1/{}~{}/?includes=openCourseAssets'.format(
            id_course, id_lecture
        )
    )


def URL_SUPPLEMENT(id_course, id_supplement):
    return (
        URL_ROOT
        + '/api/onDemandSupplements.v1/{}~{}?includes=asset&fields=openCourseAssets.v1(typeName)%2CopenCourseAssets.v1(definition)'.format(
            id_course, id_supplement
        )
    )


def URL_ASSET(ids):
    return (
        URL_ROOT
        + '/api/assets.v1?ids={}&fields=audioSourceUrls%2C+videoSourceUrls%2C+videoThumbnailUrls%2C+fileExtension%2C+tags'.format(
            ','.join(ids)
        )
    )


class DlCourseraException(Exception):
    pass


class SpecNotExistExcepton(DlCourseraException):
    def __init__(self, slug):
        super().__init__('The specialization %s does not exist' % slug)
        self.slug = slug


class CourseNotExistExcepton(DlCourseraException):
    def __init__(self, slug):
        super().__init__('The course %s does not exist' % slug)
        self.slug = slug


class NotFoundExcepton(DlCourseraException):
    def __init__(self, slug):
        super().__init__('The specialization/course %s is not found' % slug)
        self.slug = slug


class CookiesExpiredException(DlCourseraException):
    def __init__(self):
        super().__init__('The cookies.txt expired')


class BadResponseException(DlCourseraException):
    def __init__(self, d):
        super().__init__('Bad response: %s' % d)
        self.d = d


class Spec(MyDict):
    def __init__(self, *, id_=None, name=None, slug=None):
        super().__init__()

        self['type'] = 'Spec'
        self['id'] = id_
        self['name'] = name
        self['slug'] = slug
        self['courses'] = []


class Course(MyDict):
    def __init__(self, *, id_=None, name=None, slug=None):
        super().__init__()

        self['type'] = 'Course'
        self['id'] = id_
        self['name'] = name
        self['slug'] = slug
        self['modules'] = []
        self['references'] = []


class CourseReference(MyDict):
    def __init__(self, *, id_=None, name=None, slug=None, item=None):
        super().__init__()

        self['type'] = 'CourseReference'
        self['id'] = id_
        self['name'] = name
        self['slug'] = slug
        self['item'] = item


class CourseMaterialModule(MyDict):
    def __init__(self, *, id_=None, name=None, slug=None):
        super().__init__()

        self['type'] = 'Module'
        self['id'] = id_
        self['name'] = name
        self['slug'] = slug
        self['lessons'] = []


class CourseMaterialLesson(MyDict):
    def __init__(self, *, id_=None, name=None, slug=None):
        super().__init__()

        self['type'] = 'Lesson'
        self['id'] = id_
        self['name'] = name
        self['slug'] = slug
        self['items'] = []


class CourseMaterialLecture(MyDict):
    def __init__(self, *, id_=None, name=None, slug=None):
        super().__init__()

        self['type'] = 'Lecture'
        self['id'] = id_
        self['name'] = name
        self['slug'] = slug
        self['videos'] = []
        self['assets'] = []


class CourseMaterialSupplement(MyDict):
    def __init__(self, *, id_=None, name=None, slug=None):
        super().__init__()

        self['type'] = 'Supplement'
        self['id'] = id_
        self['name'] = name
        self['slug'] = slug
        self['items'] = []


class CourseMaterialSupplementItem(MyDict):
    pass


class CourseMaterialSupplementItemCML(CourseMaterialSupplementItem):
    def __init__(self, *, html, assets):
        super().__init__()

        self['type'] = 'CML'
        self['html'] = html
        self['assets'] = assets


class CourseMaterialNotebook(MyDict):
    def __init__(self, *, id_=None, name=None, slug=None):
        super().__init__()

        self['type'] = 'Notebook'
        self['id'] = id_
        self['name'] = name
        self['slug'] = slug


class Video(MyDict):
    def __init__(self, url_video, url_subtitle=None):
        super().__init__()

        self['url_video'] = url_video
        if url_subtitle is not None:
            self['url_subtitle'] = url_subtitle


class Asset(MyDict):
    def __init__(self, id_, url, name):
        super().__init__()

        self['id'] = id_
        self['url'] = url
        self['name'] = name
